#include "internal.hpp"

namespace CaDiCaL {

/*------------------------------------------------------------------------*/

Internal::Internal ()
:
  mode (SEARCH),
  unsat (false),
  iterating (false),
  preprocessing (false),
  termination_forced (false),
  stable (false),
  reported (false),
  rephased (0),
  vsize (0),
  max_var (0),
  level (0),
  vals (0),
  marks (0),
  frozentab (0),
  i2e (0),
  scinc (1.0),
  scores (this),
  stab (0),
  vtab (0),
  ltab (0),
  ftab (0),
  btab (0),
  otab (0),
  ptab (0),
  ntab (0),
  ntab2 (0),
  big (0),
  wtab (0),
  conflict (0),
  ignore (0),
  propagated (0),
  propagated2 (0),
  best_assigned (0),
  target_assigned (0),
  no_conflict_until (0),
  saved_target (0),
  saved_best (0),
  proof (0),
  checker (0),
  tracer (0),
  opts (this),
#ifndef QUIET
  profiles (this),
#endif
  arena (this),
  prefix ("c "),
  internal (this),
  external (0)
{
  control.push_back (Level (0, 0));
  binary_subsuming.redundant = false;
  binary_subsuming.size = 2;
}

Internal::~Internal () {
  for (const auto & c : clauses)
    delete_clause (c);
  if (proof) delete proof;
  if (tracer) delete tracer;
  if (checker) delete checker;
  if (vtab) DELETE_ONLY (vtab, Var, vsize);
  if (ltab) DELETE_ONLY (ltab, Link, vsize);
  if (ftab) DELETE_ONLY (ftab, Flags, vsize);
  if (btab) DELETE_ONLY (btab, long, vsize);
  if (stab) DELETE_ONLY (stab, double, vsize);
  if (ptab) DELETE_ONLY (ptab, int, 2*vsize);
  if (big) RELEASE_DELETE (big, Bins, 2*vsize);
  if (vals) { vals -= vsize; DELETE_ONLY (vals, signed_char, 2*vsize); }
  if (marks) DELETE_ONLY (marks, signed_char, vsize);
  if (phases.saved) DELETE_ONLY (phases.saved, Phase, vsize);
  if (phases.target) DELETE_ONLY (phases.target, Phase, vsize);
  if (phases.best) DELETE_ONLY (phases.best, Phase, vsize);
  if (i2e) DELETE_ONLY (i2e, int, vsize);
  if (frozentab) DELETE_ONLY (frozentab, unsigned, vsize);
  if (otab) reset_occs ();
  if (ntab) reset_noccs ();
  if (wtab) reset_watches ();
}

/*------------------------------------------------------------------------*/

void Internal::enlarge_vals (size_t new_vsize) {
  signed_char * new_vals;
  NEW_ZERO (new_vals, signed_char, 2*new_vsize);
  new_vals += new_vsize;
  if (vals) memcpy (new_vals - max_var, vals - max_var, 2*max_var + 1);
  vals -= vsize;
  DELETE_ONLY (vals, signed_char, 2*vsize);
  vals = new_vals;
}

void Internal::enlarge (int new_max_var) {
  assert (!level);
  size_t new_vsize = vsize ? 2*vsize : 1 + (size_t) new_max_var;
  while (new_vsize <= (size_t) new_max_var) new_vsize *= 2;
  LOG ("enlarge internal size from %ld to new size %ld", vsize, new_vsize);
  // Ordered in the size of allocated memory (larger block first).
  assert (!vsize || !otab);
  if (!vsize || wtab)
  ENLARGE_ZERO (wtab, Watches, 2*vsize, 2*new_vsize);
  ENLARGE_ONLY (vtab, Var, vsize, new_vsize);
  ENLARGE_ONLY (ltab, Link, vsize, new_vsize);
  ENLARGE_ZERO (btab, long, vsize, new_vsize);
  ENLARGE_ZERO (stab, double, vsize, new_vsize);
  ENLARGE_ONLY (ptab, int, 2*vsize, 2*new_vsize);
  ENLARGE_ONLY (i2e, int, vsize, new_vsize);
  enlarge_vals (new_vsize);
  ENLARGE_ZERO (frozentab, unsigned, vsize, new_vsize);
  ENLARGE_ONLY (phases.saved, Phase, vsize, new_vsize);
  ENLARGE_ZERO (phases.target, Phase, vsize, new_vsize);
  ENLARGE_ZERO (phases.best, Phase, vsize, new_vsize);
  ENLARGE_ZERO (marks, signed_char, vsize, new_vsize);
  ENLARGE_ONLY (ftab, Flags, vsize, new_vsize);
  vsize = new_vsize;
}

// Slightly different than 'bump_variable' since the variable is not
// enqueued at all.

inline void Internal::init_enqueue (int idx) {
  Link * l = ltab + idx;
  if (opts.reverse) {
    l->prev = 0;
    if (queue.first) {
      assert (!ltab[queue.first].prev);
      ltab[queue.first].prev = idx;
      btab[idx] = btab[queue.first] - 1;
    } else {
      assert (!queue.last);
      queue.last = idx;
      btab[idx] = 0;
    }
    assert (btab[idx] <= stats.bumped);
    l->next = queue.first;
    queue.first = idx;
    if (!queue.unassigned)
      update_queue_unassigned (queue.last);
  } else {
    l->next = 0;
    if (queue.last) {
      assert (!ltab[queue.last].next);
      ltab[queue.last].next = idx;
    } else {
      assert (!queue.first);
      queue.first = idx;
    }
    btab[idx] = ++stats.bumped;
    l->prev = queue.last;
    queue.last = idx;
    update_queue_unassigned (queue.last);
  }
}

// Initialize VMTF queue from current 'old_max_var + 1' to 'new_max_var'.
// This incorporates an initial variable order.  We currently simply assume
// that variables with smaller index are more important.  This is the same
// as in MiniSAT (implicitly) and also matches the 'scores' initialization.
//
void Internal::init_queue (int old_max_var, int new_max_var) {
  LOG ("initializing VMTF queue from %d to %d",
    old_max_var + 1, new_max_var);
  assert (old_max_var < new_max_var);
  assert (!level);
  for (int idx = old_max_var; idx < new_max_var; idx++)
    init_enqueue (idx + 1);
}

// This initializes variables on the binary 'scores' heap also with
// smallest variable index first (thus picked first) and larger indices at
// the end.
//
void Internal::init_scores (int old_max_var, int new_max_var) {
  LOG ("initializing EVSIDS scores from %d to %d",
    old_max_var + 1, new_max_var);
  for (int i = old_max_var + 1; i <= new_max_var; i++)
    scores.push_back (i);
}

void Internal::init (int new_max_var) {
  if (new_max_var <= max_var) return;
  if (level) backtrack ();
  LOG ("initializing %d internal variables from %d to %d",
    new_max_var - max_var, max_var + 1, new_max_var);
  if ((size_t) new_max_var >= vsize) enlarge (new_max_var);
  signed_char val = opts.phase ? 1 : -1;
  for (int i = max_var + 1; i <= new_max_var; i++)
    phases.saved[i] = val;
#ifndef NDEBUG
  for (int i = -new_max_var; i < -max_var; i++) assert (!vals[i]);
  for (int i = max_var + 1; i <= new_max_var; i++) assert (!vals[i]);
  for (int i = max_var + 1; i <= new_max_var; i++) assert (!frozentab[i]);
  for (int i = max_var + 1; i <= new_max_var; i++) assert (!marks[i]);
  for (int i = max_var + 1; i <= new_max_var; i++) assert (!btab[i]);
#endif
  for (int i = 2*(max_var + 1); i <= 2*new_max_var+1; i++) ptab[i] = -1;
  for (int i = max_var + 1; i <= new_max_var; i++) ftab[i].init ();
  assert (!btab[0]);
  int old_max_var = max_var;
  max_var = new_max_var;
  init_queue (old_max_var, new_max_var);
  init_scores (old_max_var, new_max_var);
  int initialized = new_max_var - old_max_var;
  stats.vars += initialized;
  stats.unused += initialized;
  stats.inactive += initialized;
  LOG ("finished initializing %d internal variables", initialized);
}

void Internal::add_original_lit (int lit) {
  assert (abs (lit) <= max_var);
  if (lit) {
    original.push_back (lit);
  } else {
    if (proof) proof->add_original_clause (original);
    add_new_original_clause ();
    original.clear ();
  }
}

/*------------------------------------------------------------------------*/

// This is the main CDCL loop with interleaved inprocessing.

int Internal::search () {

  int res = 0;

  START (search);
  assert (!stable);
  START (unstable);

  while (!res) {
         if (unsat) res = 20;
    else if (!propagate ()) analyze ();      // propagate and analyze
    else if (iterating) iterate ();          // report learned unit
    else if (satisfied ()) res = 10;         // found model
    else if (terminating ()) break;          // limit hit or sync abort
    else if (restarting ()) restart ();      // restart by backtracking
    else if (rephasing ()) rephase ();       // reset variable phases
    else if (reducing ()) reduce ();         // collect useless clauses
    else if (probing ()) probe ();           // failed literal probing
    else if (subsuming ()) subsume ();       // subsumption algorithm
    else if (eliminating ()) elim ();        // variable elimination
    else if (compacting ()) compact ();      // collect variables
    else res = decide ();                    // next decision
  }

  if (stable) STOP (stable);
  else        STOP (unstable);
  STOP (search);

  return res;
}

/*------------------------------------------------------------------------*/

void Internal::init_limits () {

  const bool incremental = lim.initialized;
  if (incremental) LOG ("reinitializing limits incrementally");
  else LOG ("initializing limits and increments");

  const char * mode = 0;
  (void) mode;

  if (incremental) mode = "keeping";
  else {
    last.reduce.conflicts = -1;
    lim.reduce = stats.conflicts + opts.reduceint;
    mode = "initial";
  }
  LOG ("%s reduce limit %ld after %ld conflicts",
    mode, lim.reduce, lim.reduce - stats.conflicts);

  if (incremental) mode = "keeping";
  else {
    lim.flush = opts.flushinit;
    inc.flush = opts.flushinit;
    mode = "initial";
  }
  LOG ("%s flush limit %ld interval %ld",
    mode, lim.flush, inc.flush);

  if (incremental) mode = "keeping";
  else {
    lim.subsume = stats.conflicts + opts.subsumeint;
    mode = "initial";
  }
  LOG ("%s subsume limit %ld after %ld conflicts",
    mode, lim.subsume, lim.subsume - stats.conflicts);

  if (incremental) mode = "keeping";
  else {
    last.elim.marked = -1;
    lim.elim = stats.conflicts + opts.elimint;
    mode = "initial";
  }
  LOG ("%s elim limit %ld after %ld conflicts",
    mode, lim.elim, lim.elim - stats.conflicts);

  if (incremental) mode = "keeping";
  else {
  lim.elimbound = min (opts.elimboundmin, opts.elimboundmax);
  mode = "initial";
  }
  LOG ("%s elimination bound %ld",
    mode, lim.elimbound);

  if (incremental) mode = "keeping";
  else {
    lim.probe = stats.conflicts + opts.probeint;
    mode = "initial";
  }
  LOG ("%s probe limit %ld after %ld conflicts",
    mode, lim.probe, lim.probe - stats.conflicts);

  if (!incremental) {

    last.ternary.marked = -1;

    lim.compact = stats.conflicts + opts.compactint;
    LOG ("initial compact limit %ld increment %ld",
      lim.compact, lim.compact - stats.conflicts);
  }

  lim.rephase = stats.conflicts + opts.rephaseint;
  lim.rephased[0] = lim.rephased[1] = 0;
  LOG ("new rephase limit %ld after %ld conflicts",
    lim.rephase, lim.rephase - stats.conflicts);

  lim.restart = stats.conflicts + opts.restartint;
  LOG ("new restart limit %ld increment %ld",
    lim.restart, lim.restart - stats.conflicts);

  reported = false;
  lim.report = 0;

  if (!incremental) {
    assert (!stable);
    init_averages ();
  } else if (stable) {
    swap_averages ();
    stable = false;
  }

  inc.stabilize = opts.stabilizeint;
  lim.stabilize = stats.conflicts + inc.stabilize;
  LOG ("new stabilize limit %ld after %ld conflicts",
    lim.stabilize, inc.stabilize);

  if (opts.stabilize && opts.reluctant) {
    LOG ("new restart reluctant doubling sequence period %d",
      opts.reluctant);
    reluctant.enable (opts.reluctant, opts.reluctantmax);
  } else reluctant.disable ();

  if (inc.conflicts < 0) {
    lim.conflicts = -1;
    LOG ("no limit on conflicts");
  } else {
    lim.conflicts = stats.conflicts + inc.conflicts;
    LOG ("conflict limit after %ld conflicts at %ld conflicts",
      inc.conflicts, lim.conflicts);
  }

  if (inc.decisions < 0) {
    lim.decisions = -1;
    LOG ("no limit on decisions");
  } else {
    lim.decisions = stats.decisions + inc.decisions;
    LOG ("conflict limit after %ld decisions at %ld decisions",
      inc.decisions, lim.decisions);
  }

  if (inc.preprocessing <= 0) {
    lim.preprocessing = 0;
    LOG ("no preprocessing");
  } else {
    lim.preprocessing = inc.preprocessing;
    LOG ("limiting to %d preprocessing rounds", lim.preprocessing);
  }

  lim.initialized = true;
}

bool Internal::preprocess (int round) {
  (void) round;
  if (unsat) return false;
  if (!max_var) return false;
  START (preprocess);
  struct { int vars; long clauses; } before, after;
  before.vars = active ();
  before.clauses = stats.current.irredundant;
  stats.preprocessings++;
  assert (!preprocessing);
  preprocessing = true;
  PHASE ("preprocessing", stats.preprocessings,
    "starting round %d with %d variables and %ld clauses",
    round, before.vars, before.clauses);
  int old_elimbound = lim.elimbound;
  if (opts.probe) probe (false);
  if (opts.elim) elim (false);
  after.vars = active ();
  after.clauses = stats.current.irredundant;
  assert (preprocessing);
  preprocessing = false;
  PHASE ("preprocessing", stats.preprocessings,
    "finished round %d with %d variables and %ld clauses",
    round, after.vars, after.clauses);
  STOP (preprocess);
  report ('o');
  if (unsat) return false;
  if (after.vars < before.vars) return true;
  if (old_elimbound < lim.elimbound) return true;
  return false;
}

int Internal::solve () {
  assert (clause.empty ());
  START (solve);
  if (level) backtrack ();
  int res;
  if (unsat) {
    LOG ("already inconsistent");
    res = 20;
  } else if (!propagate ()) {
    LOG ("root level propagation produces conflict");
    learn_empty_clause ();
    res = 20;
  } else {

    init_limits ();

    if (external->taintedlits.empty ()) {
      LOG ("no tainted literals and nothing to restore");
      report ('*');
    } else {
      report ('+');
      external->restore_clauses ();
      internal->report ('r');
    }

    if (opts.simplify)
      for (int i = 0; i < lim.preprocessing; i++)
        if (!preprocess (i))
          break;

    if (unsat) {
      LOG ("inconsistent after preprocessing");
      res = 20;
    } else if (terminating ()) res = 0;
    else {
      assert (!stable);
      report ('{');
      res = search ();
      report (stable ? ']' : '}');
    }
  }
  if (termination_forced) {
    termination_forced = false;
    LOG ("reset forced termination");
  }
       if (res == 10) report ('1');
  else if (res == 20) report ('0');
  else                report ('?');
  STOP (solve);
  return res;
}

/*------------------------------------------------------------------------*/

void Internal::print_stats () {
  stats.print (this);
  if (checker) checker->print_stats ();
}

/*------------------------------------------------------------------------*/

// Only useful for debugging purposes.

void Internal::dump (Clause * c) {
  for (const auto & lit : *c)
    printf ("%d ", lit);
  printf ("0\n");
}

void Internal::dump () {
  long m = assumptions.size ();
  for (int idx = 1; idx <= max_var; idx++)
    if (fixed (idx)) m++;
  for (const auto & c : clauses)
    if (!c->garbage) m++;
  printf ("p cnf %d %ld\n", max_var, m);
  for (int idx = 1; idx <= max_var; idx++) {
    const int tmp = fixed (idx);
    if (tmp) printf ("%d 0\n", tmp < 0 ? -idx : idx);
  }
  for (const auto & c : clauses)
    if (!c->garbage) dump (c);
  for (const auto & lit : assumptions)
    printf ("%d 0\n", lit);
  fflush (stdout);
}

/*------------------------------------------------------------------------*/

bool Internal::traverse_clauses (ClauseIterator & it) {
  vector<int> eclause;
  if (unsat) return it.clause (eclause);
  for (const auto & c : clauses) {
    if (c->garbage) continue;
    if (c->redundant) continue;
    bool satisfied = false;
    for (const auto & ilit : *c) {
      const int tmp = fixed (ilit);
      if (tmp > 0) { satisfied = true; break; }
      if (tmp < 0) continue;
      const int elit = externalize (ilit);
      eclause.push_back (elit);
    }
    if (!satisfied && !it.clause (eclause))
      return false;
    eclause.clear ();
  }
  return true;
}

}
