#include "internal.hpp"

namespace CaDiCaL {

/*------------------------------------------------------------------------*/

bool Internal::rephasing () {
  if (!opts.rephase) return false;
  return stats.conflicts > lim.rephase;
}

/*------------------------------------------------------------------------*/

// These are the different strategies for resetting saved variable phases.

char Internal::rephase_original () {
  stats.rephasedorig++;
  signed char val = opts.phase ? 1 : -1;        // original = initial
  PHASE ("rephase", stats.rephased,
    "switching to original phase %d", val);
  for (int idx = 1; idx <= max_var; idx++)
    phases.saved[idx] = val;
  return 'O';
}

char Internal::rephase_flipping () {
  stats.rephasedflip++;
  PHASE ("rephase", stats.rephased,
    "flipping all phases individually");
  for (int idx = 1; idx <= max_var; idx++)
    phases.saved[idx] *= -1;
  return 'F';
}

char Internal::rephase_inverted () {
  stats.rephasedinv++;
  signed char val = opts.phase ? -1 : 1;
  PHASE ("rephase", stats.rephased,
    "switching to inverted original phase %d", val);
  for (int idx = 1; idx <= max_var; idx++)
    phases.saved[idx] = val;
  return 'I';
}

char Internal::rephase_random () {
  stats.rephasedrand++;
  PHASE ("rephase", stats.rephased,
    "resetting all phases randomly");
  Random random (opts.seed);                    // global seed
  random += stats.rephased;                     // different every time
  for (int idx = 1; idx <= max_var; idx++)
    phases.saved[idx] = random.generate_bool () ? -1 : 1;
  return '#';
}

// Best phases are those at the largest trail without conflict. See the
// corresponding code and comments in 'update_target_and_best'.

char Internal::rephase_best () {
  stats.rephasedbest++;
  PHASE ("rephase", stats.rephased,
    "overwriting best phases as saved phases");
  signed char val;
  for (int idx = 1; idx <= max_var; idx++)
    if ((val = phases.best[idx]))
      phases.saved[idx] = val;
  return 'B';
}

/*------------------------------------------------------------------------*/

void Internal::rephase () {

  stats.rephased++;
  PHASE ("rephase", stats.rephased,
    "reached rephase limit %ld after %ld conflicts",
    lim.rephase, stats.conflicts);

  // Report current 'target' and 'best' and then set 'rephased' below, which
  // will trigger reporting the new 'target' and 'best' after updating it in
  // the next 'update_target_and_best' called from the next 'backtrack'.
  //
  report ('~', 1);

  backtrack ();

  if (stable) {
    memset (phases.target, 0, (max_var + 1) * sizeof (Phase));
    PHASE ("rephase", stats.rephased, "reset all target phases");
  }
  target_assigned = 0;

  size_t count = lim.rephased[stable]++;
  char type;

  if (!opts.stabilize) {
    // (inverted,best,flipping,best,random,best,original,best)^\omega
    switch (count % 8) {
      case 0: type = rephase_inverted (); break;
      case 2: type = rephase_flipping (); break;
      case 4: type = rephase_random (); break;
      case 6: type = rephase_original (); break;
      default: type = rephase_best (); break;
    }
  } else if (stable) {
    // original,inverted,(best,original,best,inverted)^\omega
    if (!count) type = rephase_original ();
    else if (count == 1) type = rephase_inverted ();
    else switch (count % 4) {
      case 1: type = rephase_inverted (); break;
      case 3: type = rephase_original (); break;
      default: type = rephase_best (); break;
    }
  } else {
    // flipping,(random,best,flipping,best)^\omega
    if (!count) type = rephase_flipping ();
    else switch (count % 4) {
      case 1: type = rephase_random (); break;
      case 3: type = rephase_flipping (); break;
      default: type = rephase_best (); break;
    }
  }

  long delta = opts.rephaseint * (stats.rephased + 1);
  lim.rephase = stats.conflicts + delta;

  PHASE ("rephase", stats.rephased,
    "new rephase limit %ld after %ld conflicts",
    lim.rephase, delta);

  // This will trigger to report the effect of this new set of phases at the
  // 'backtrack' (actually 'update_target_and_best') after the next
  // conflict, as well as resetting 'best_assigned' then to allow to compute
  // a new "best" assignment at that point.
  //
  last.rephase.conflicts = stats.conflicts;
  rephased = type;
}

}
